import pytest

from spectate import mvc
from spectate.mvc import Undefined as undef

from .mock import model_events, events_to_comparable_list


_method_call_and_expected_event = [
    {
        "value": {},
        "method": "__setattr__",
        "args": ["a", 1],
        "kwargs": {},
        "events": [{"old": undef, "new": 1, "attr": "a"}],
    },
    {
        "value": {"a": None},
        "method": "__setattr__",
        "args": ["a", 1],
        "kwargs": {},
        "events": [{"old": None, "new": 1, "attr": "a"}],
    },
    {
        "value": {"a": 1},
        "method": "__delattr__",
        "args": ["a"],
        "kwargs": {},
        "events": [{"old": 1, "new": undef, "attr": "a"}],
    },
]


@pytest.mark.parametrize("expectation", _method_call_and_expected_event)
def test_basic_events(expectation):
    value, actual_events = model_events(mvc.Object, expectation["value"])
    method = getattr(value, expectation["method"])
    args = expectation.get("args", [])
    kwargs = expectation.get("kwargs", {})
    method(*args, **kwargs)
    expected_events = expectation["events"]
    assert events_to_comparable_list(actual_events) == events_to_comparable_list(
        expected_events
    )



# SEE END OF FILE FOR LICENSE

import inspect
import itertools

from .utils import Sentinel
from .base import Model, Control


__all__ = ["Structure", "List", "Dict", "Set", "Object", "Undefined"]


Undefined = Sentinel("Undefined")


class Structure(Model):
    def _notify_model_views(self, events):
        for e in events:
            if "new" in e:
                new = e["new"]
                if isinstance(new, Model):
                    self._attach_child_model(new)
            if "old" in e:
                old = e["old"]
                if isinstance(old, Model):
                    self._remove_child_model(old)
        super()._notify_model_views(events)


class List(Structure, list):
    """A :mod:`spectate.mvc` enabled ``list``."""

    _control_setitem = Control(
        "__setitem__", before="_control_before_setitem", after="_control_after_setitem"
    )

    _control_delitem = Control(
        "__delitem__", before="_control_before_delitem", after="_control_after_delitem"
    )

    _control_insert = Control(
        "insert", before="_control_before_insert", after="_control_after_insert"
    )

    _control_append = Control("append", after="_control_after_append")

    _control_extend = Control(
        "__init__, extend",
        before="_control_before_extend",
        after="_control_after_extend",
    )

    _control_pop = Control(
        "pop", before="_control_before_pop", after="_control_after_delitem"
    )

    _control_clear = Control(
        "clear", before="_control_before_clear", after="_control_after_clear"
    )

    _control_remove = Control(
        "remove", before="_control_before_remove", after="_control_after_delitem"
    )

    _control_rearrangement = Control(
        "sort, reverse",
        before="_control_before_rearrangement",
        after="_control_after_rearrangement",
    )

    def _control_before_setitem(self, call, notify):
        index = call["args"][0]
        try:
            old = self[index]
        except KeyError:
            old = Undefined
        return index, old

    def _control_after_setitem(self, answer, notify):
        index, old = answer["before"]
        new = self[index]
        if new is not old:
            notify(index=index, old=old, new=new)

    def _control_before_delitem(self, call, notify):
        index = call["args"][0]
        return index, self[index:]

    def _control_after_delitem(self, answer, notify):
        index, old = answer["before"]
        for i, x in enumerate(old):
            try:
                new = self[index + i]
            except IndexError:
                new = Undefined
            notify(index=(i + index), old=x, new=new)

    def _control_before_insert(self, call, notify):
        index = call["args"][0]
        return index, self[index:]

    def _control_after_insert(self, answer, notify):
        index, old = answer["before"]
        for i in range(index, len(self)):
            try:
                o = old[i]
            except IndexError:
                o = Undefined
            notify(index=i, old=o, new=self[i])

    def _control_after_append(self, answer, notify):
        notify(index=len(self) - 1, old=Undefined, new=self[-1])

    def _control_before_extend(self, call, notify):
        return len(self)

    def _control_after_extend(self, answer, notify):
        for i in range(answer["before"], len(self)):
            notify(index=i, old=Undefined, new=self[i])

    def _control_before_pop(self, call, notify):
        if not call["args"]:
            index = len(self) - 1
        else:
            index = call["args"][0]
        return index, self[index:]

    def _control_before_clear(self, call, notify):
        return self.copy()

    def _control_after_clear(self, answer, notify):
        for i, v in enumerate(answer["before"]):
            notify(index=i, old=v, new=Undefined)

    def _control_before_remove(self, call, notify):
        index = self.index(call["args"][0])
        return index, self[index:]

    def _control_before_rearrangement(self, call, notify):
        return self.copy()

    def _control_after_rearrangement(self, answer, notify):
        old = answer["before"]
        for i, v in enumerate(old):
            if v != self[i]:
                notify(index=i, old=v, new=self[i])


class Dict(Structure, dict):
    """A :mod:`spectate.mvc` enabled ``dict``."""

    _control_setitem = Control(
        "__setitem__, setdefault",
        before="_control_before_setitem",
        after="_control_after_setitem",
    )

    _control_delitem = Control(
        "__delitem__, pop",
        before="_control_before_delitem",
        after="_control_after_delitem",
    )

    _control_update = Control(
        "__init__, update",
        before="_control_before_update",
        after="_control_after_update",
    )

    _control_clear = Control(
        "clear", before="_control_before_clear", after="_control_after_clear"
    )

    def _control_before_setitem(self, call, notify):
        key = call["args"][0]
        old = self.get(key, Undefined)
        return key, old

    def _control_after_setitem(self, answer, notify):
        key, old = answer["before"]
        new = self[key]
        if new != old:
            notify(key=key, old=old, new=new)

    def _control_before_delitem(self, call, notify):
        key = call["args"][0]
        try:
            return key, self[key]
        except KeyError:
            # the base method will error on its own
            pass

    def _control_after_delitem(self, answer, notify):
        key, old = answer["before"]
        notify(key=key, old=old, new=Undefined)

    def _control_before_update(self, call, notify):
        if len(call["args"]):
            args = call["args"][0]
            if inspect.isgenerator(args):
                # copy generator so it doesn't get exhausted
                args = itertools.tee(args)[1]
            new = dict(args)
            new.update(call["kwargs"])
        else:
            new = call["kwargs"]
        old = {k: self.get(k, Undefined) for k in new}
        return old

    def _control_after_update(self, answer, notify):
        for k, v in answer["before"].items():
            if self[k] != v:
                notify(key=k, old=v, new=self[k])

    def _control_before_clear(self, call, notify):
        return self.copy()

    def _control_after_clear(self, answer, notify):
        for k, v in answer["before"].items():
            notify(key=k, old=v, new=Undefined)


class Set(Structure, set):
    """A :mod:`spectate.mvc` enabled ``set``."""

    _control_update = Control(
        [
            "__init__",
            "clear",
            "update",
            "difference_update",
            "intersection_update",
            "add",
            "pop",
            "remove",
            "symmetric_difference_update",
            "discard",
        ],
        before="_control_before_update",
        after="_control_after_update",
    )

    def _control_before_update(self, call, notify):
        return self.copy()

    def _control_after_update(self, answer, notify):
        new = self.difference(answer["before"])
        old = answer["before"].difference(self)
        if new or old:
            notify(new=new, old=old)


class Object(Structure):
    """A :mod:`spectat.mvc` enabled ``object``."""

    _control_attr_change = Control(
        "__setattr__, __delattr__",
        before="_control_before_attr_change",
        after="_control_after_attr_change",
    )

    def __init__(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            setattr(self, k, v)

    def _control_before_attr_change(self, call, notify):
        return call["args"][0], getattr(self, call["args"][0], Undefined)

    def _control_after_attr_change(self, answer, notify):
        attr, old = answer["before"]
        new = getattr(self, attr, Undefined)
        if new != old:
            notify(attr=attr, old=old, new=new)



import pytest

from spectate import mvc

from .mock import model_events, Counter


def test_hold_events():
    counter, events = model_events(Counter)

    with mvc.hold(counter) as cache:
        counter.increment(1)
        assert cache == [{"old": 0, "new": 1}]

        counter.increment(1)
        assert cache == [{"old": 0, "new": 1}, {"old": 1, "new": 2}]

        # Pop off one of the events so
        # it isn't sent to notifiers.
        cache.pop()

    assert events == [{"old": 0, "new": 1}]


def test_hold_uses_events_from_reducer():
    counter, events = model_events(Counter)

    def reducer(model, events):
        assert events == [{"old": 0, "new": 1}]
        yield {"custom": "event-1"}
        yield {"custom": "event-2"}

    with mvc.hold(counter, reducer=reducer):
        counter.increment(1)

    assert events == [{"custom": "event-1"}, {"custom": "event-2"}]


def test_rollback_events():
    counter, events = model_events(Counter)

    with pytest.raises(ValueError):
        with mvc.rollback(counter):
            counter.increment(1)
            raise ValueError()

    assert not events


def test_rollback_calls_undo_without_side_effects():
    calls = []
    counter, events = model_events(Counter)

    def undo(model, events, error):
        calls.append(1)
        assert error is error_from_rollback
        assert events == ({"old": 0, "new": 1},)
        # this decrement should not notify
        model.decrement(1)

    with pytest.raises(ValueError):
        with mvc.rollback(counter, undo=undo):
            counter.increment(1)
            error_from_rollback = ValueError()
            raise error_from_rollback

    assert calls
    assert counter.value == 0


def test_mute_events():
    counter, events = model_events(Counter)

    with mvc.mute(counter):
        counter.increment(1)
        counter.increment(1)

    assert events == []



"""A modules which exports specate's Model-View-Controller utilities in a common namespace

For more info:

- :mod:`spectate.base`
- :mod:`spectate.events`
- :mod:`spectate.models`
"""

from .base import *  # noqa
from .events import *  # noqa
from .models import *  # noqa



from spectate import mvc


def events_to_comparable_list(events):
    return tuple(sorted(tuple(sorted((k, evt[k]) for k in evt)) for evt in events))


def model_events(model, *args, **kwargs):
    cached_events = []

    if not isinstance(model, mvc.Model):
        model = model(*args, **kwargs)

    @mvc.view(model)
    def cache(model, events):
        cached_events.extend(events)

    return model, cached_events


class Counter(mvc.Model):
    def __init__(self):
        self.value = 0

    def increment(self, amount):
        self.value += amount

    def decrement(self, amount):
        self.value -= amount

    _control_change = mvc.Control(
        ["increment", "decrement"],
        before="_control_before_change",
        after="_control_after_change",
    )

    def _control_before_change(self, call, notify):
        return self.value

    def _control_after_change(self, answer, notify):
        notify(old=answer["before"], new=self.value)



import pytest

from spectate import mvc

from .mock import model_events, events_to_comparable_list


_method_call_and_expected_event = [
    {
        "value": {1, 2, 3},
        "method": "clear",
        "args": [],
        "kwargs": {},
        "events": [{"old": {1, 2, 3}, "new": set()}],
    },
    {
        "value": set(),
        "method": "update",
        "args": [[1, 2, 3]],
        "kwargs": {},
        "events": [{"old": set(), "new": {1, 2, 3}}],
    },
    {
        "value": set(),
        "method": "add",
        "args": [1],
        "kwargs": {},
        "events": [{"old": set(), "new": {1}}],
    },
    {
        "value": {1},
        "method": "remove",
        "args": [1],
        "kwargs": {},
        "events": [{"old": {1}, "new": set()}],
    },
    {
        "value": {1},
        "method": "pop",
        "args": [],
        "kwargs": {},
        "events": [{"old": {1}, "new": set()}],
    },
    {
        "value": {1},
        "method": "discard",
        "args": [1],
        "kwargs": {},
        "events": [{"old": {1}, "new": set()}],
    },
    {"value": {1}, "method": "discard", "args": [2], "kwargs": {}, "events": []},
    {
        "value": {1, 2, 3},
        "method": "intersection_update",
        "args": [{2, 3, 4}],
        "kwargs": {},
        "events": [{"old": {1}, "new": set()}],
    },
    {
        "value": {1, 2, 3},
        "method": "symmetric_difference_update",
        "args": [{2, 3, 4}],
        "kwargs": {},
        "events": [{"old": {2, 3}, "new": {4}}],
    },
]


@pytest.mark.parametrize("expectation", _method_call_and_expected_event)
def test_basic_events(expectation):
    value, actual_events = model_events(mvc.Set, expectation["value"])
    method = getattr(value, expectation["method"])
    args = expectation.get("args", [])
    kwargs = expectation.get("kwargs", {})
    method(*args, **kwargs)
    expected_events = expectation["events"]
    assert events_to_comparable_list(actual_events) == events_to_comparable_list(
        expected_events
    )



from inspect import signature
from functools import wraps
from typing import (
    Any,
    Union,
    Callable,
    Optional,
    TypeVar,
    Tuple,
    Dict,
    List,
    Iterator,
    overload,
)
from contextlib import contextmanager
from weakref import WeakValueDictionary


__all__ = ["Model", "Control", "view", "unview", "views", "link", "unlink", "notifier"]

Event = Dict[str, Any]
TupleOfEvents = Tuple[Event, ...]
ViewFunction = Callable[["Model", TupleOfEvents], None]


def views(model: "Model") -> List[ViewFunction]:
    """Return a model's views keyed on what events they respond to.

    Model views are added by calling :func:`view` on a model.
    """
    if not isinstance(model, Model):
        raise TypeError("Expected a Model, not %r." % model)
    return model._model_views[:]


_F = TypeVar("_F", bound=ViewFunction)


@overload
def view(model: "Model") -> Callable[[_F], _F]:
    ...


@overload
def view(model: "Model", function: ViewFunction) -> None:
    ...


def view(
    model: "Model", function: Optional[ViewFunction] = None
) -> Optional[Callable[[_F], _F]]:
    """A decorator for registering a callback to a model

    Parameters:
        model: the model object whose changes the callback should respond to.

    Examples:
        .. code-block:: python

            from spectate import mvc

            items = mvc.List()

            @mvc.view(items)
            def printer(items, events):
                for e in events:
                    print(e)

            items.append(1)
    """
    if not isinstance(model, Model):
        raise TypeError("Expected a Model, notself._model_notifier() %r." % model)

    def setup(function: _F) -> _F:
        model._attach_model_view(function)
        return function

    if function is not None:
        setup(function)
        return None
    else:
        return setup


def unview(model: "Model", function: ViewFunction) -> None:
    """Remove a view callbcak from a model.

    Parameters:
        model: The model which contains the view function.
        function: The callable which was registered to the model as a view.

    Raises:
        ValueError: If the given ``function`` is not a view of the given ``model``.
    """
    model._remove_model_view(function)


def link(source: "Model", *targets: "Model") -> None:
    """Attach all of the source's present and future view functions to the targets.

    Parameters:
        source: The model whose view functions will be attached to the targets.
        targets: The models that will acquire the source's view functions.
    """
    for t in targets:
        source._attach_child_model(t)


def unlink(source: "Model", *targets: "Model") -> None:
    """Remove all of the source's present and future view functions from the targets.

    Parameters:
        source: The model whose view functions will be removed from the targets.
        targets: The models that will no longer share view functions with the source.
    """
    for t in targets:
        source._remove_child_model(t)


@contextmanager
def notifier(model: "Model") -> Iterator[Callable[..., None]]:
    """Manually send notifications to the given model.

    Parameters:
        model: The model whose views will recieve notifications

    Returns:
        A function whose keyword arguments become event data.

    Example:

        .. code-block:: python

            m = Model()

            @view(m)
            def printer(m, events):
                for e in events:
                    print(e)

            with notifier(m) as notify:
                # the view should print out this event
                notify(x=1, y=2)
    """
    events = []

    def notify(*args, **kwargs):
        events.append(dict(*args, **kwargs))

    yield notify

    if events:
        model._notify_model_views(tuple(events))


class Control:
    """An object used to define control methods on a :class:`Model`

    A "control" method on a :class:`Model` is one which reacts to another method being
    called. For example there is a control method on the
    :class:`~spectate.mvc.models.List`
    which responds when :meth:`~spectate.mvc.models.List.append` is called.

    A control method is a slightly modified :ref:`beforeback <Control Beforebacks>` or
    :ref:`afterback <Control Afterbacks>` that accepts an extra ``notify`` argument.
    These are added to a control object by calling :meth:`Control.before` or
    :meth:`Control.after` respectively. The ``notify`` arugment is a function which
    allows a control method to send messages to :func:`views <view>` that are registered
    to a :class:`Model`.

    Parameters:
        methods:
            The names of the methods on the model which this control will react to
            When they are calthrough the Nodeled. This is either a comma seperated
            string, or a list of strings.
        before:
            A control method that reacts before any of the given ``methods`` are
            called. If given as a callable, then that function will be used as the
            callback. If given as a string, then the control will look up a method
            with that name when reacting (useful when subclassing).
        after:
            A control method that reacts after any of the given ``methods`` are
            alled. If given as a callable, then that function will be used as the
            callback. If given as a string, then the control will look up a method
            with that name when reacting (useful when subclassing).

    Examples:
        Control methods are registered to a :class:`Control` with a ``str`` or function.
        A string may refer to the name of a method on a `Model` while a function should
        be decorated under the same name as the :class:`Control` object to preserve the
        namespace.

        .. code-block:: python

            from spectate import mvc

            class X(mvc.Model):

                _control_method = mvc.Control("method").before("_control_before_method")

                def _control_before_method(self, call, notify):
                    print("before")

                # Note how the method uses the same name. It
                # would be redundant to use a different one.
                @_control_a.after
                def _control_method(self, answer, notify):
                    print("after")

                def method(self):
                    print("during")

            x = X()
            x.method()

        .. code-block:: text

            before
            during
            after
    """

    def __init__(
        self,
        methods: Union[list, tuple, str],
        *,
        before: Union[Callable, str] = None,
        after: Union[Callable, str] = None,
    ):
        if isinstance(methods, (list, tuple)):
            self.methods = tuple(methods)
        elif isinstance(methods, str):
            self.methods = tuple(map(str.strip, methods.split(",")))
        else:
            raise ValueError("methods must be a string or list of strings")
        self.name = None
        if isinstance(before, Control):
            before = before._before
        self._before = before
        if isinstance(after, Control):
            after = after._after
        self._after = after

    def __get__(self, obj, cls):
        if obj is None:
            return self
        else:
            return BoundControl(obj, self)

    def __set_name__(self, cls, name):
        if not issubclass(cls, Model):
            raise TypeError("Can only define a control on a Model, not %r" % cls)
        if self.name:
            msg = "Control was defined twice - %r and %r."
            raise RuntimeError(msg % (self.name, name))
        else:
            self.name = name
        for m in self.methods:
            setattr(cls, m, self._create_controlled_method(cls, m))

    def _create_controlled_method(self, cls, name):
        method = getattr(cls, name)

        @wraps(method)
        def wrapped_method(obj, *args, **kwargs):
            cls = type(obj)
            bound_control = self.__get__(obj, cls)

            before_control = bound_control.before
            if before_control is not None:
                before_value = before_control(
                    obj, {"name": name, "args": args, "kwargs": kwargs}
                )
            else:
                before_value = None

            result = method.__get__(obj, cls)(*args, **kwargs)

            after_control = bound_control.after
            if after_control is not None:
                after_control(
                    obj, {"before": before_value, "name": name, "value": result}
                )

            return result

        return wrapped_method


class BoundControl:
    def __init__(self, obj, ctrl):
        self._obj = obj
        self._cls = type(obj)
        self._name = ctrl.name
        self._before = ctrl._before
        self._after = ctrl._after
        self.methods = ctrl.methods

    @property
    def before(self):
        if self._before is None:
            method_name = self._name + "_before"
            if hasattr(self._obj, method_name):
                before = getattr(self._obj, method_name)
            else:
                return None
        else:
            before = self._before

        if isinstance(before, str):
            before = getattr(self._obj, before)
        elif hasattr(before, "__get__"):
            before = before.__get__(self._obj, type(self._obj))

        @wraps(before)
        def beforeback(value, call):
            def parameters():
                meth = getattr(value, call["name"])
                bound = signature(meth).bind(*call["args"], **call["kwargs"])
                return dict(bound.arguments)

            with notifier(value) as notify:
                return before(dict(call, parameters=parameters), notify)

        return beforeback

    @property
    def after(self):
        if self._after is None:
            return None
        else:
            after = self._after

        if isinstance(after, str):
            after = getattr(self._obj, after)
        elif hasattr(after, "__get__"):
            after = after.__get__(self._obj, type(self._obj))

        @wraps(after)
        def afterback(value, answer):
            with notifier(value) as notify:
                return after(answer, notify)

        return afterback


class Model:
    """An object that can be :class:`controlled <Control>` and :func:`viewed <view>`.

    Users should define :class:`Control` methods and then :func:`view` the change
    events those controls emit. This process starts by defining controls on a subclass
    of :class:`Model`.

    Examples:
        .. code-block:: python

            from specate import mvc

            class Object(Model):

                _control_attr_change = Control(
                    "__setattr__, __delattr__",
                    before="_control_before_attr_change",
                    after="_control_after_attr_change",
                )

                def __init__(self, *args, **kwargs):
                    for k, v in dict(*args, **kwargs).items():
                        setattr(self, k, v)

                def _control_before_attr_change(self, call, notify):
                    return call["args"][0], getattr(self, call["args"][0], Undefined)

                def _control_after_attr_change(self, answer, notify):
                    attr, old = answer["before"]
                    new = getattr(self, attr, Undefined)
                    if new != old:
                        notify(attr=attr, old=old, new=new)

            o = Object()

            @mvc.view(o)
            def printer(o, events):
                for e in events:
                    print(e)
    """

    _model_views: List[ViewFunction]
    _inner_models: "WeakValueDictionary[int, Model]"

    def __new__(cls, *args: Any, **kwargs: Any) -> "Model":
        new = super().__new__
        if new is not object.__new__:
            self = new(cls, *args, **kwargs)  # type: ignore
        else:
            self = new(cls)

        object.__setattr__(self, "_model_views", [])
        object.__setattr__(self, "_inner_models", WeakValueDictionary())

        return self

    def _attach_child_model(self, model: "Model") -> None:
        self._inner_models[id(model)] = model
        for v in self._model_views:
            model._attach_model_view(v)

    def _remove_child_model(self, model: "Model") -> None:
        try:
            del self._inner_models[id(model)]
        except KeyError:
            pass
        else:
            for v in self._model_views:
                model._remove_model_view(v)

    def _attach_model_view(self, function: ViewFunction) -> None:
        self._model_views.append(function)
        for inner in self._inner_models.values():
            inner._attach_model_view(function)

    def _remove_model_view(self, function: ViewFunction) -> None:
        self._model_views.remove(function)
        for inner in self._inner_models.values():
            inner._remove_model_view(function)

    def _notify_model_views(self, events: TupleOfEvents):
        for view in self._model_views:
            view(self, events)



from pytest import raises

from spectate import mvc

from .mock import model_events, Counter


def test_control_is_only_for_model():
    """Controls must be defined on a Model"""
    with raises(RuntimeError):

        class X:
            _control = mvc.Control("something")


def test_control_using_functions():

    calls = []

    def _control_before(value, call, notify):
        calls.append("before")

    def _control_after(value, call, notify):
        calls.append("after")

    class X(mvc.Model):
        def method(self):
            calls.append("during")

        control = mvc.Control("method", before=_control_before, after=_control_after)

    X().method()
    assert calls == ["before", "during", "after"]


def test_control_using_string_reference():

    calls = []

    class X(mvc.Model):
        def method(self):
            pass

        control = mvc.Control("method", before="before", after="after")

        def before(self, call, notify):
            calls.append("before")

        def after(self, answer, notify):
            calls.append("after")

    X().method()
    assert calls == ["before", "after"]


def test_control_events_were_sent():
    counter, events = model_events(Counter)

    counter.increment(1)
    counter.decrement(2)
    counter.increment(3)
    counter.decrement(4)

    assert events == [
        {"old": 0, "new": 1},
        {"old": 1, "new": -1},
        {"old": -1, "new": 2},
        {"old": 2, "new": -2},
    ]


def test_override_control_methods_in_subclass():
    class MyCounter(Counter):
        def _control_before_change(self, call, notify):
            notify(message="before")
            return super()._control_before_change(call, notify)

    counter, events = model_events(MyCounter)

    counter.increment(1)

    assert events == [{"message": "before"}, {"old": 0, "new": 1}]


def test_add_new_control_in_subclass():
    class MyCounter(Counter):

        _added_control = mvc.Control("increment", before="_added_before")

        def _added_before(self, call, notify):
            notify(message="before")

    counter, events = model_events(MyCounter)

    counter.increment(1)

    assert events == [{"message": "before"}, {"old": 0, "new": 1}]


def test_structure_events():
    class Container(mvc.Structure):
        def __init__(self, name):
            self.name = name
            self.value = None

        def set(self, value):
            self.value = value

        _control_changes = mvc.Control(
            "set", before="_before_change", after="_after_change"
        )

        def _before_change(self, call, notify):
            return self.value

        def _after_change(self, answer, notify):
            notify(old=answer["before"], new=self.value)

        def __repr__(self):
            return "Container(%r)" % self.name

    s0 = Container("s0")
    s1 = Container("s1")
    s2 = Container("s2")

    calls = []

    @mvc.view(s0)
    def on_change(value, events):
        calls.append(value)

    s0.set(s1)
    s1.set(s2)
    s2.set(None)

    assert calls == [s0, s1, s2]
    calls.clear()

    # remove children
    s0.set(None)

    assert calls == [s0]
    calls.clear()

    # these should not trigger events because they are no longer
    # attached to the root container s0
    s1.set(None)
    s2.set(None)

    assert not calls


def test_notifier_context_manager():
    calls = []
    m = mvc.Model()

    @mvc.view(m)
    def viewer(m, events):
        calls.extend(events)

    with mvc.notifier(m) as notify:
        notify(data=1)
        assert calls == []

    assert calls == [{"data": 1}]
    calls.clear()

    with mvc.notifier(m) as notify:
        notify(data=1)
        notify(data=2)

    assert calls == [{"data": 1}, {"data": 2}]


def test_link_and_unlink_inner_models():
    calls = []

    parent = mvc.Model()
    child = mvc.Model()
    grandchild = mvc.Model()

    mvc.link(parent, child)
    mvc.link(child, grandchild)

    def trigger_events():
        with mvc.notifier(grandchild) as notify:
            notify({"data": 1})
        with mvc.notifier(child) as notify:
            notify({"data": 2})
        with mvc.notifier(parent) as notify:
            notify({"data": 3})
        copy = calls[:]
        calls.clear()
        return copy

    @mvc.view(parent)
    def viewer(value, events):
        calls.append({"v": value, "e": list(events)})

    assert trigger_events() == [
        {"v": grandchild, "e": [{"data": 1}]},
        {"v": child, "e": [{"data": 2}]},
        {"v": parent, "e": [{"data": 3}]},
    ]

    mvc.unlink(child, grandchild)
    assert trigger_events() == [
        {"v": child, "e": [{"data": 2}]},
        {"v": parent, "e": [{"data": 3}]},
    ]

    mvc.unlink(parent, child)
    assert trigger_events() == [{"v": parent, "e": [{"data": 3}]}]


def test_unlink_middleman_stops_view_of_leaf_models():
    calls = []

    parent = mvc.Model()
    child = mvc.Model()
    grandchild = mvc.Model()

    mvc.link(parent, child)
    mvc.link(child, grandchild)

    def trigger_events():
        with mvc.notifier(grandchild) as notify:
            notify({"data": 1})
        with mvc.notifier(child) as notify:
            notify({"data": 2})
        with mvc.notifier(parent) as notify:
            notify({"data": 3})
        copy = calls[:]
        calls.clear()
        return copy

    @mvc.view(parent)
    def viewer(value, events):
        calls.append({"v": value, "e": list(events)})

    mvc.unlink(parent, child)

    assert trigger_events() == [{"v": parent, "e": [{"data": 3}]}]



import pytest

from spectate import mvc
from spectate.mvc import Undefined as undef

from .mock import model_events, events_to_comparable_list


_method_call_and_expected_event = [
    {
        "value": [1, 2, 3],
        "method": "__setitem__",
        "args": [1, 5],
        "kwargs": {},
        "events": [{"old": 2, "new": 5, "index": 1}],
    },
    {
        "value": [1, 2, 3],
        "method": "__delitem__",
        "args": [0],
        "kwargs": {},
        "events": [
            {"old": 1, "new": 2, "index": 0},
            {"old": 2, "new": 3, "index": 1},
            {"old": 3, "new": undef, "index": 2},
        ],
    },
    {
        "value": [1, 2, 3],
        "method": "append",
        "args": [4],
        "kwargs": {},
        "events": [{"old": undef, "new": 4, "index": 3}],
    },
    {
        "value": [1, 2, 3],
        "method": "pop",
        "args": [],
        "kwargs": {},
        "events": [{"old": 3, "new": undef, "index": 2}],
    },
    {
        "value": [1, 2, 3],
        "method": "pop",
        "args": [0],
        "kwargs": {},
        "events": [
            {"old": 1, "new": 2, "index": 0},
            {"old": 2, "new": 3, "index": 1},
            {"old": 3, "new": undef, "index": 2},
        ],
    },
    {
        "value": [2, 3],
        "method": "insert",
        "args": [0, 1],
        "kwargs": {},
        "events": [
            {"old": 2, "new": 1, "index": 0},
            {"old": 3, "new": 2, "index": 1},
            {"old": undef, "new": 3, "index": 2},
        ],
    },
    {
        "value": [],
        "method": "extend",
        "args": [[1, 2, 3]],
        "kwargs": {},
        "events": [
            {"old": undef, "new": 1, "index": 0},
            {"old": undef, "new": 2, "index": 1},
            {"old": undef, "new": 3, "index": 2},
        ],
    },
    {
        "value": [],
        "method": "extend",
        "args": [(i for i in range(1, 4))],
        "kwargs": {},
        "events": [
            {"old": undef, "new": 1, "index": 0},
            {"old": undef, "new": 2, "index": 1},
            {"old": undef, "new": 3, "index": 2},
        ],
    },
    {
        "value": [1, 2, 3],
        "method": "clear",
        "args": [],
        "kwargs": {},
        "events": [
            {"old": 1, "new": undef, "index": 0},
            {"old": 2, "new": undef, "index": 1},
            {"old": 3, "new": undef, "index": 2},
        ],
    },
    {"value": [], "method": "clear", "args": [], "kwargs": {}, "events": []},
    {
        "value": [1, 2, 3],
        "method": "remove",
        "args": [1],
        "kwargs": {},
        "events": [
            {"old": 1, "new": 2, "index": 0},
            {"old": 2, "new": 3, "index": 1},
            {"old": 3, "new": undef, "index": 2},
        ],
    },
    {
        "value": [3, 2, 1],
        "method": "sort",
        "args": [],
        "kwargs": {},
        "events": [{"old": 3, "new": 1, "index": 0}, {"old": 1, "new": 3, "index": 2}],
    },
    {
        "value": [1, 2, 3],
        "method": "reverse",
        "args": [],
        "kwargs": {},
        "events": [{"old": 1, "new": 3, "index": 0}, {"old": 3, "new": 1, "index": 2}],
    },
]


@pytest.mark.parametrize("expectation", _method_call_and_expected_event)
def test_basic_events(expectation):
    value, actual_events = model_events(mvc.List, expectation["value"])
    method = getattr(value, expectation["method"])
    args = expectation.get("args", [])
    kwargs = expectation.get("kwargs", {})
    method(*args, **kwargs)
    expected_events = expectation["events"]
    assert events_to_comparable_list(actual_events) == events_to_comparable_list(
        expected_events
    )



from contextlib import contextmanager
from typing import Iterator, Callable, Optional, List

from .base import Model, Event, TupleOfEvents

__all__ = ["hold", "mute", "rollback"]


_EventReducerFunc = Callable[[Model, List[Event]], List[Event]]


@contextmanager
def hold(
    model: Model, reducer: Optional[_EventReducerFunc] = None
) -> Iterator[List[Event]]:
    """Temporarilly withold change events in a modifiable list.

    All changes that are captured within a "hold" context are forwarded to a list
    which is yielded to the user before being sent to views of the given ``model``.
    If desired, the user may modify the list of events before the context is left in
    order to change the events that are ultimately sent to the model's views.

    Parameters:
        model:
            The model object whose change events will be temporarilly witheld.
        reducer:
            A function for modifying the events list at the end of the context.
            Its signature is ``(model, events) -> new_events`` where ``model`` is the
            given model, ``events`` is the complete list of events produced in the
            context, and the returned ``new_events`` is a list of events that will
            actuall be distributed to views.

    Notes:
        All changes witheld from views will be sent as a single notification. For
        example if you view a :class:`specate.mvc.models.List` and its ``append()``
        method is called three times within a :func:`hold` context,


    Examples:
        Note how the event from ``l.append(1)`` is omitted from the printed statements.

        .. code-block:: python

            from spectate import mvc

            l = mvc.List()

            mvc.view(d, lambda d, e: list(map(print, e)))

            with mvc.hold(l) as events:
                l.append(1)
                l.append(2)

                del events[0]

        .. code-block:: text

            {'index': 1, 'old': Undefined, 'new': 2}
    """
    if not isinstance(model, Model):
        raise TypeError("Expected a Model, not %r." % model)
    events: List[Event] = []
    restore = model.__dict__.get("_notify_model_views")
    model._notify_model_views = lambda e: events.extend(e)  # type: ignore

    try:
        yield events
    finally:
        if restore is None:
            del model._notify_model_views
        else:
            model._notify_model_views = restore  # type: ignore

        if reducer is not None:
            events = reducer(model, events)

        model._notify_model_views(tuple(events))


@contextmanager
def rollback(
    model: Model,
    undo: Callable[[Model, TupleOfEvents, Exception], None] = None,
    reducer: Optional[_EventReducerFunc] = None,
) -> Iterator[None]:
    """Withold events if an error occurs.

    Generall operate

    Parameters:
        model:
            The model object whose change events may be witheld.
        undo:
            An optional function for reversing any changes that may have taken place.
            Its signature is ``(model, events, error)`` where ``model`` is the given
            model, ``events`` is a list of all the events that took place, and ``error``
            is the exception that was riased. Any changes that you make to the model
            within this function will not produce events.

    Examples:

        Simple supression of events:

        .. code-block:: python

            from spectate import mvc

            d = mvc.Dict()

            @mvc.view(d)
            def should_not_be_called(d, events):
                # we never call this view
                assert False

            try:
                with mvc.rollback(d):
                    d["a"] = 1
                    d["b"]  # key doesn't exist
            except KeyError:
                pass

        Undo changes for a dictionary:

        .. code-block:: python

            from spectate import mvc

            def undo_dict_changes(model, events, error):
                seen = set()
                for e in reversed(events):
                    if e.old is mvc.Undefined:
                        del model[e.key]
                    else:
                        model[e.key] = e.old

            try:
                with mvc.rollback(d, undo=undo_dict_changes):
                    d["a"] = 1
                    d["b"] = 2
                    print(d)
                    d["c"]
            except KeyError:
                pass
            print(d)

        .. code-block:: python

            {'a': 1, 'b': 2}
            {}
    """
    with hold(model, reducer=reducer) as events:
        try:
            yield None
        except Exception as error:
            if undo is not None:
                with mute(model):
                    undo(model, tuple(events), error)
            events.clear()
            raise


@contextmanager
def mute(model: Model) -> Iterator[None]:
    """Block a model's views from being notified.

    All changes within a "mute" context will be blocked. No content is yielded to the
    user as in :func:`hold`, and the views of the model are never notified that changes
    took place.

    Parameters:
        mode: The model whose change events will be blocked.

    Examples:

        The view is never called due to the :func:`mute` context:

        .. code-block:: python

            from spectate import mvc

            l = mvc.List()

            @mvc.view(l)
            def raises(events):
                raise ValueError("Events occured!")

            with mvc.mute(l):
                l.append(1)
    """
    if not isinstance(model, Model):
        raise TypeError("Expected a Model, not %r." % model)
    restore = model.__dict__.get("_notify_model_views")
    model._notify_model_views = lambda e: None  # type: ignore
    try:
        yield None
    finally:
        if restore is None:
            del model._notify_model_views
        else:
            model._notify_model_views = restore  # type: ignore
